"""Defines the CloseableQueue class and the Close exception class."""

from queue import Empty, Full
from time import monotonic as time
import queue as _Queue

class Closed(Exception):
    """Exception raised by CloseableQueue.put/get on a closed queue."""
    pass

def CloseableQueueFactory(base=_Queue.Queue, name="CloseableQueue"):
    """Create a closeable descendant class of `base`.

    Used instead of a mixin approach because the Queue module's classes
      are old-style.
    """
    class CloseableQueue(base):
        """This class provides a means to permanently close a queue.

        Attempts to `put` to a closed queue will raise the `Closed` exception.

        Attempts to `get` from an *empty* closed queue will raise the same.

        Blocked `put`s and `get`s on a queue which is subsequently closed
          will also raise the `Closed` exception under the same circumstances.

        A queue can be closed either by calling its `close` method
          or by passing `last=True` to an invocation of `put`.

        If the latter is done, the entire operation is performed atomically;
          the close will only take place if the put succeeds.
        """
        def __init__(self, *args, **kwargs):
            base.__init__(self, *args, **kwargs)
            assert not hasattr(self, '_closed')
            self._closed = False

        def close(self):
            """Close the queue.

            This will prevent further `put`s, and only allow `get`s
              until the contents are depleted.

            `put`s and `get`s which are prevented raise `Closed`.

            Calling `close` will also cause `Closed` exceptions to be raised
              in blocked `get`s or `put`s as though they had just been called.

            Normally it is only useful to call this method
              from a thread which is the sole producer or sole consumer.
            """
            self.mutex.acquire()
            try:
                if not self._closed:
                    self._closed = True
                    self.not_empty.notify_all()
                    self.not_full.notify_all()
            finally:
                self.mutex.release()

        def closed(self):
            """True iff the queue is closed.  Unreliable like `empty` and `full`."""
            # Probably not necessary to use a protected section here,
            #   but better safe than bug-ridden.
            # This also leaves the door open
            #   for more complex implementations of `_closed` as a property.
            self.mutex.acquire()
            n = self._closed
            self.mutex.release()
            return n

        def put(self, item, block=True, timeout=None, last=False):
            """Put an item into the queue.

            Works as does `Queue.Queue.put`, but with these differences:

            If the queue is closed, raises Closed.

            If `last` is True and the put succeeds,
              the queue will be atomically closed.

            Also raises `Closed` in the event that the queue is closed
              while the `put` is blocked.
            """
            self.not_full.acquire()
            try:
                if self.maxsize > 0:
                    if not block:
                        if self._qsize() == self.maxsize and not self._closed:
                            raise Full
                    elif timeout is None:
                        while self._qsize() == self.maxsize and not self._closed:
                            self.not_full.wait()
                    elif timeout < 0:
                        raise ValueError("'timeout' must be a positive number")
                    else:
                        endtime = _time() + timeout
                        while self._qsize() == self.maxsize and not self._closed:
                            remaining = endtime - _time()
                            if remaining <= 0.0:
                                raise Full
                            self.not_full.wait(remaining)
                if self._closed:
                    raise Closed
                self._put(item)
                self.unfinished_tasks += 1
                if last:
                    self._closed = True
                    self.not_empty.notify_all()
                    self.not_full.notify_all()
                else:
                    self.not_empty.notify()
            finally:
                self.not_full.release()

        def get(self, block=True, timeout=None):
            """Remove and return an item from the queue.

            Works as does `Queue.Queue.get` except that
              a `get` on a closed queue will raise `Closed`.

            Similarly, a blocked `get` will raise `Closed`
              if the queue is closed during the block.
            """
            self.not_empty.acquire()
            try:
                if not block:
                    if not self._qsize() and not self._closed:
                        raise Empty
                elif timeout is None:
                    while not self._qsize() and not self._closed:
                        self.not_empty.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a positive number")
                else:
                    endtime = _time() + timeout
                    while not self._qsize() and not self._closed:
                        remaining = endtime - _time()
                        if remaining <= 0.0:
                            raise Empty
                        self.not_empty.wait(remaining)
                if self._closed and not self._qsize():
                    raise Closed
                item = self._get()
                self.not_full.notify()
                return item
            finally:
                self.not_empty.release()
    CloseableQueue.__name__ = name
    return CloseableQueue

CloseableQueue = CloseableQueueFactory()
CloseableLifoQueue = CloseableQueueFactory(_Queue.LifoQueue,
                                           "CloseableLifoQueue")
CloseablePriorityQueue = CloseableQueueFactory(_Queue.PriorityQueue,
                                               "CloseablePriorityQueue")

def dequeue(q, getargs={}, on_empty='stop'):
    """Generates values from the queue `q`.

    This is a fairly flexible function which can also be meaningfully applied
      to non-closeable queues, by passing a `timeout` in `getargs`.

    The default arguments are suitable for generating an iteration
      from a queue which is closed after/as the last value is `put` to it.

    To avoid false StopIterations, if a timeout is used with a closeable queue,
      `on_empty` should be passed as `raise` or some sentinel value.

    `getargs` is a dict which will comprise the keyword arguments to `q.get`.

    `on_empty` determines the response to an `Empty` exception raised by `get`.
      If 'raise', the `Empty` exception will be raised.
      If 'stop', the `Empty` exception will be treated as a `Closed` exception.
      Otherwise, the value passed in `on_empty` will be yielded.

    Note that with the default value of `getargs`,
      `Empty` exceptions will not occur.
    """
    try:
        while True:
            yield q.get(**getargs)
    except Closed:
        raise StopIteration
    except Empty:
        if on_empty == 'raise':
            raise
        elif on_empty == 'stop':
            raise StopIteration
        else:
            yield on_empty

def enqueue(it, q, putargs={}, join=False, close=True):
    """`put`s the successive values of the iterable `it` into `q`.

    The default values will close the queue after the final iterated value.

    `putargs` is a dict which will comprise the keyword arguments to `q.put`.

    If `close` is true,
      `q` must support the `close` method, i.e. be a CloseableQueue.
    This will have the effect of closing the queue after the end of iteration.

    If `join` is true, the queue is joined after the values are put,
      and after optionally being closed.
    """
    for value in iter(it):
        q.put(value, **putargs)
    if close:
        q.close()
    if join:
        q.join()

def EnqueueThread(it, q=None, name='enqueue', start=True, enqueue=enqueue,
                  **kwargs):
    """Starts a thread which enqueues the values of the iterable `it`.

    If a queue is not passed, a new one is created.

    A reference to the queue is stored as the property `q`
      of the returned thread.

    The thread will be started unless `start` is false.

    This is ideal for processing generators
      and other iterables which lazily process I/O-bound operations.

    Passing the return value of a `dequeue` call will have the effect
      of creating a processing thread which pulls from and pushes to
      thread-safe data structures.

    Additional keyword arguments are passed on to `enqueue`.
    """
    from threading import Thread
    if q is None:
        q = CloseableQueue()
    thread = Thread(name=name,
                    target=enqueue,
                    args=(it, q),
                    kwargs=kwargs)
    thread.q = q
    if start:
        thread.start()
    return thread



import multiprocessing
import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _utils import eprint

def send(self, value):
	if self._server == None:
		self._pipe[1].send(value)
	else:
		session_id = self._get_session_id()
		self._request(session_id, value=value)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process_send(self, request):
	try:
		self._pipe[0].send(request["data"]["value"])
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def recv(self):
	if self._server == None:
		return self._pipe[1].recv()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process_recv(self, request):
	try:
		self._respond_ok(request["session_id"], value=self._pipe[0].recv())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _locals_pipe_send(self, name, value):
	if self._server == None:
		if name not in self._actual_pipes:
			self._actual_pipes[name] = multiprocessing.Pipe()

		self._actual_pipes[name][1].send(value)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, value=value)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process__locals_pipe_send(self, request):
	try:
		name = request["data"]["name"]
		if name not in self._actual_pipes:
			self._actual_pipes[name] = multiprocessing.Pipe()

		self._actual_pipes[name][0].send(request["data"]["value"])
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _locals_pipe_recv(self, name):
	if self._server == None:
		if name not in self._actual_pipes:
			self._actual_pipes[name] = multiprocessing.Pipe()

		return self._actual_pipes[name][1].recv()
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__locals_pipe_recv(self, request):
	try:
		name = request["data"]["name"]
		if name not in self._actual_pipes:
			self._actual_pipes[name] = multiprocessing.Pipe()

		self._respond_ok(request["session_id"], value=self._actual_pipes[name][0].recv())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _locals_pipe_close(self, name):
	if self._server == None:
		if name not in self._actual_pipes:
			return

		try:
			self._actual_pipes[name][0].close()
		except:
			pass

		try:
			self._actual_pipes[name][1].close()
		except:
			pass
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process__locals_pipe_close(self, request):
	try:
		name = request["data"]["name"]
		if name not in self._actual_pipes:
			self._respond_ok(request["session_id"])
			return

		try:
			self._actual_pipes[name][0].close()
		except:
			pass

		try:
			self._actual_pipes[name][1].close()
		except:
			pass

		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _locals_pipes_delitem(self, name):
	if self._server == None:
		if name not in self._actual_pipes:
			return

		try:
			self._actual_pipes[name][0].close()
		except:
			pass

		try:
			self._actual_pipes[name][1].close()
		except:
			pass

		del self._actual_pipes[name]
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process__locals_pipes_delitem(self, request):
	try:
		name = request["data"]["name"]
		if name not in self._actual_pipes:
			self._respond_ok(request["session_id"])
			return

		try:
			self._actual_pipes[request["data"]["name"]][0].close()
		except:
			pass

		try:
			self._actual_pipes[request["data"]["name"]][1].close()
		except:
			pass

		del self._actual_pipes[request["data"]["name"]]
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _locals_pipes_clear(self):
	if self._server == None:
		for name in self._actual_pipes:
			try:
				self._actual_pipes[name][0].close()
			except:
				pass

			try:
				self._actual_pipes[name][1].close()
			except:
				pass

		self._actual_pipes.clear()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process__locals_pipes_clear(self, request):
	try:
		for name in self._actual_pipes:
			try:
				self._actual_pipes[name][0].close()
			except:
				pass

			try:
				self._actual_pipes[name][1].close()
			except:
				pass

		self._actual_pipes.clear()
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _locals_pipes_len(self):
	if self._server == None:
		return len(self._actual_pipes)
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["len"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__locals_pipes_len(self, request):
	try:
		self._respond_ok(request["session_id"], len=self._actual_pipes.__len__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _locals_pipes_contains(self, name):
	if self._server == None:
		return self._actual_pipes.__contains__(name)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["contains"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__locals_pipes_contains(self, request):
	try:
		self._respond_ok(request["session_id"], contains=self._actual_pipes.__contains__(request["data"]["name"]))
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _locals_pipes_keys(self):
	if self._server == None:
		return self._actual_pipes.keys()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["keys"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__locals_pipes_keys(self, request):
	try:
		self._respond_ok(request["session_id"], keys=self._actual_pipes.keys())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _locals_pipes_iter(self):
	if self._server == None:
		return self._actual_pipes.__iter__()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["iter"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__locals_pipes_iter(self, request):
	try:
		self._respond_ok(request["session_id"], iter=self._actual_pipes.__iter__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)


import sys
import os

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _utils import eprint
from _NodeInternalClasses import CloseableQueue

def put(self, value, timeout=None, block=True):
	if self._server == None:
		self._queue.put(value)
	else:
		session_id = self._get_session_id()
		self._request(session_id, value=value, timeout=timeout, block=block)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def get(self, timeout=None, block=True):
	if self._server == None:
		return self._queue.get()
	else:
		session_id = self._get_session_id()
		self._request(session_id, timeout=timeout, block=block)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def qsize(self):
	if self._server == None:
		return self._queue.qsize()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["qsize"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queue_put(self, name, value, timeout, block, is_private):
	if self._server == None:
		if not is_private:
			if name not in self._actual_queues:
				self._actual_queues[name] = CloseableQueue()
			self._actual_queues[name].put(value, timeout=timeout, block=block)
		else:
			if name not in self._actual_result_queues_for_future:
				self._actual_result_queues_for_future[name] = CloseableQueue()
			self._actual_result_queues_for_future[name].put(value, timeout=timeout, block=block)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, value=value, timeout=timeout, block=block, is_private=is_private)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queue_get(self, name, timeout, block, is_private):
	if self._server == None:
		if not is_private:
			if name not in self._actual_queues:
				self._actual_queues[name] = CloseableQueue()
			return self._actual_queues[name].get(timeout=timeout, block=block)
		else:
			if name not in self._actual_result_queues_for_future:
				self._actual_result_queues_for_future[name] = CloseableQueue()
			return self._actual_result_queues_for_future[name].get(timeout=timeout, block=block)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, timeout=timeout, block=block, is_private=is_private)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queue_close(self, name, is_private):
	if self._server == None:
		if not is_private:
			if name not in self._actual_queues:
				return
			self._actual_queues[name].close()
		else:
			if name not in self._actual_result_queues_for_future:
				return
			self._actual_result_queues_for_future[name].close()
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, is_private=is_private)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queue_len(self, name, is_private):
	if self._server == None:
		if not is_private:
			if name not in self._actual_queues:
				return 0
			else:
				return self._actual_queues[name].qsize()
		else:
			if name not in self._actual_result_queues_for_future:
				return 0
			else:
				return self._actual_result_queues_for_future[name].qsize()
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, is_private=is_private)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["len"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queues_delitem(self, name, is_private):
	if self._server == None:
		if not is_private:
			self._actual_queues[name].close()
			del self._actual_queues[name]
		else:
			self._actual_result_queues_for_future[name].close()
			del self._actual_result_queues_for_future[name]
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, is_private=is_private)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queues_len(self, is_private):
	if self._server == None:
		if not is_private:
			return self._actual_queues.__len__()
		else:
			return self._actual_result_queues_for_future.__len__()
	else:
		session_id = self._get_session_id()
		self._request(session_id, is_private=is_private)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["len"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queues_contains(self, name, is_private):
	if self._server == None:
		if not is_private:
			return self._actual_queues.__contains__(name)
		else:
			return self._actual_result_queues_for_future.__contains__(name)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, is_private=is_private)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["contains"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queues_keys(self, is_private):
	if self._server == None:
		if not is_private:
			return self._actual_queues.keys()
		else:
			return self._actual_result_queues_for_future.keys()
	else:
		session_id = self._get_session_id()
		self._request(session_id, is_private=is_private)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["keys"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queues_iter(self, is_private):
	if self._server == None:
		if not is_private:
			return self._actual_queues.__iter__()
		else:
			return self._actual_result_queues_for_future.__iter__()
	else:
		session_id = self._get_session_id()
		self._request(session_id, is_private=is_private)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["iter"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _locals_queues_clear(self, is_private):
	if self._server == None:
		if not is_private:
			for name in self._actual_queues:
				self._actual_queues[name].close()
			self._actual_queues.clear()
		else:
			for name in self._actual_result_queues_for_future:
				self._actual_result_queues_for_future[name].close()
			self._actual_result_queues_for_future.clear()
	else:
		session_id = self._get_session_id()
		self._request(session_id, is_private=is_private)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process_put(self, request):
	try:
		self._queue.put(request["data"]["value"], timeout=request["data"]["timeout"], block=request["block"])
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process_get(self, request):
	try:
		value = self._queue.get(timeout=request["data"]["timeout"], block=request["block"])
		self._respond_ok(request["session_id"], value=value)
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process_qsize(self, request):
	try:
		qsize = self._queue.qsize()
		self._respond_ok(request["session_id"], qsize=qsize)
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queue_put(self, request):
	try:
		name = request["data"]["name"]
		if not request["data"]["is_private"]:
			if name not in self._actual_queues:
				self._actual_queues[name] = CloseableQueue()

			self._actual_queues[name].put(request["data"]["value"], timeout=request["data"]["timeout"], block=request["block"])
		else:
			if name not in self._actual_result_queues_for_future:
				self._actual_result_queues_for_future[name] = CloseableQueue()

			self._actual_result_queues_for_future[name].put(request["data"]["value"], timeout=request["data"]["timeout"], block=request["block"])

		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queue_get(self, request):
	try:
		name = request["data"]["name"]
		if not request["data"]["is_private"]:
			if name not in self._actual_queues:
				self._actual_queues[name] = CloseableQueue()

			value = self._actual_queues[request["data"]["name"]].get(timeout=request["data"]["timeout"], block=request["block"])
		else:
			if name not in self._actual_result_queues_for_future:
				self._actual_result_queues_for_future[name] = CloseableQueue()

			value = self._actual_result_queues_for_future[request["data"]["name"]].get(timeout=request["data"]["timeout"], block=request["block"])
		self._respond_ok(request["session_id"], value=value)
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queue_len(self, request):
	try:
		if not request["data"]["is_private"]:
			if request["data"]["name"] not in self._actual_queues:
				self._respond_ok(request["session_id"], len=0)
			else:
				self._respond_ok(request["session_id"], len=self._actual_queues[request["data"]["name"]].qsize())
		else:
			if request["data"]["name"] not in self._actual_result_queues_for_future:
				self._respond_ok(request["session_id"], len=0)
			else:
				self._respond_ok(request["session_id"], len=self._actual_result_queues_for_future[request["data"]["name"]].qsize())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queue_close(self, request):
	try:
		name = request["data"]["name"]
		if not request["data"]["is_private"]:
			try:
				self._actual_queues[name].close()
			except:
				pass
		else:
			try:
				self._actual_result_queues_for_future[name].close()
			except:
				pass
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queues_delitem(self, request):
	try:
		name = request["data"]["name"]
		if not request["data"]["is_private"]:
			try:
				self._actual_queues[name].close()
			except:
				pass

			del self._actual_queues[name]
		else:
			try:
				self._actual_result_queues_for_future[name].close()
			except:
				pass
			del self._actual_result_queues_for_future[name]
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queues_len(self, request):
	try:
		if not request["data"]["is_private"]:
			self._respond_ok(request["session_id"], len=self._actual_queues.__len__())
		else:
			self._respond_ok(request["session_id"], len=self._actual_result_queues_for_future.__len__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queues_contains(self, request):
	try:
		if not request["data"]["is_private"]:
			self._respond_ok(request["session_id"], contains=self._actual_queues.__contains__(request["data"]["name"]))
		else:
			self._respond_ok(request["session_id"], contains=self._actual_result_queues_for_future.__contains__(request["data"]["name"]))
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queues_keys(self, request):
	try:
		if not request["data"]["is_private"]:
			self._respond_ok(request["session_id"], keys=self._actual_queues.keys())
		else:
			self._respond_ok(request["session_id"], keys=self._actual_result_queues_for_future.keys())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queues_iter(self, request):
	try:
		if not request["data"]["is_private"]:
			self._respond_ok(request["session_id"], iter=self._actual_queues.__iter__())
		else:
			self._respond_ok(request["session_id"], iter=self._actual_result_queues_for_future.__iter__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__locals_queues_clear(self, request):
	try:
		if not request["data"]["is_private"]:
			for name in self._actual_queues:
				try:
					self._actual_queues[name].close()
				except:
					pass
			self._actual_queues.clear()
		else:
			for name in self._actual_result_queues_for_future:
				try:
					self._actual_result_queues_for_future[name].close()
				except:
					pass
			self._actual_result_queues_for_future.clear()

		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)


from __future__ import print_function
import hashlib
import sys
import os

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def md5(file_name):
	if not os.path.isfile(file_name):
		return None
	else:
		m = hashlib.md5()
		file = open(file_name, 'rb')
		while True:
			data = file.read(4096)
			if not data:
				break
			m.update(data)
		file.close()

		return m.hexdigest()

def file_size(file_name):
	if not os.path.isfile(file_name):
		return 0
	else:
		return os.path.getsize(file_name)


import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
import _NodeBasicMethods
import _NodeLocalPipesMethods
import _NodeLocalQueuesMethods
import _NodeLocalDictMethods
import _NodeGlobalQueuesMethods
import _NodeGlobalDictMethods
import _NodeFileTransferMethods
import _NodeRPCMethods

class Node:		                   
	## methods
	# basic
	__init__ = _NodeBasicMethods.__init__
	__del__ = _NodeBasicMethods.__del__
	close = _NodeBasicMethods.close
	hold_on = _NodeBasicMethods.hold_on
	_start = _NodeBasicMethods._start
	_get_session_id = _NodeBasicMethods._get_session_id
	_traceback = _NodeBasicMethods._traceback
	_process__get_session_id = _NodeBasicMethods._process__get_session_id
	_process_close = _NodeBasicMethods._process_close
	_send = _NodeBasicMethods._send
	_send_signal = _NodeBasicMethods._send_signal
	_request = _NodeBasicMethods._request
	_respond_ok = _NodeBasicMethods._respond_ok
	_respond_exception = _NodeBasicMethods._respond_exception
	_put_result = _NodeBasicMethods._put_result
	_put_exception = _NodeBasicMethods._put_exception
	_make_signal = _NodeBasicMethods._make_signal
	_recv_signal = _NodeBasicMethods._recv_signal
	_recv_response = _NodeBasicMethods._recv_response
	_recving_loop = _NodeBasicMethods._recving_loop

	# local pipes
	send = _NodeLocalPipesMethods.send
	recv = _NodeLocalPipesMethods.recv
	_locals_pipe_send = _NodeLocalPipesMethods._locals_pipe_send
	_locals_pipe_recv = _NodeLocalPipesMethods._locals_pipe_recv
	_locals_pipe_close = _NodeLocalPipesMethods._locals_pipe_close
	_locals_pipes_len = _NodeLocalPipesMethods._locals_pipes_len
	_locals_pipes_contains = _NodeLocalPipesMethods._locals_pipes_contains
	_locals_pipes_keys = _NodeLocalPipesMethods._locals_pipes_keys
	_locals_pipes_iter = _NodeLocalPipesMethods._locals_pipes_iter
	_locals_pipes_delitem = _NodeLocalPipesMethods._locals_pipes_delitem
	_locals_pipes_clear = _NodeLocalPipesMethods._locals_pipes_clear
	_process_send = _NodeLocalPipesMethods._process_send
	_process_recv = _NodeLocalPipesMethods._process_recv
	_process__locals_pipe_send = _NodeLocalPipesMethods._process__locals_pipe_send
	_process__locals_pipe_recv = _NodeLocalPipesMethods._process__locals_pipe_recv
	_process__locals_pipe_close = _NodeLocalPipesMethods._process__locals_pipe_close
	_process__locals_pipes_len = _NodeLocalPipesMethods._process__locals_pipes_len
	_process__locals_pipes_contains = _NodeLocalPipesMethods._process__locals_pipes_contains
	_process__locals_pipes_keys = _NodeLocalPipesMethods._process__locals_pipes_keys
	_process__locals_pipes_iter = _NodeLocalPipesMethods._process__locals_pipes_iter
	_process__locals_pipes_delitem = _NodeLocalPipesMethods._process__locals_pipes_delitem
	_process__locals_pipes_clear = _NodeLocalPipesMethods._process__locals_pipes_clear

	# local queues
	put = _NodeLocalQueuesMethods.put
	get = _NodeLocalQueuesMethods.get
	qsize = _NodeLocalQueuesMethods.qsize
	_locals_queue_put = _NodeLocalQueuesMethods._locals_queue_put
	_locals_queue_get = _NodeLocalQueuesMethods._locals_queue_get
	_locals_queue_close = _NodeLocalQueuesMethods._locals_queue_close
	_locals_queue_len = _NodeLocalQueuesMethods._locals_queue_len
	_locals_queues_len = _NodeLocalQueuesMethods._locals_queues_len
	_locals_queues_contains = _NodeLocalQueuesMethods._locals_queues_contains
	_locals_queues_keys = _NodeLocalQueuesMethods._locals_queues_keys
	_locals_queues_iter = _NodeLocalQueuesMethods._locals_queues_iter
	_locals_queues_delitem = _NodeLocalQueuesMethods._locals_queues_delitem
	_locals_queues_clear = _NodeLocalQueuesMethods._locals_queues_clear
	_process_put = _NodeLocalQueuesMethods._process_put
	_process_get = _NodeLocalQueuesMethods._process_get
	_process_qsize = _NodeLocalQueuesMethods._process_qsize
	_process__locals_queue_put = _NodeLocalQueuesMethods._process__locals_queue_put
	_process__locals_queue_get = _NodeLocalQueuesMethods._process__locals_queue_get
	_process__locals_queue_len = _NodeLocalQueuesMethods._process__locals_queue_len
	_process__locals_queue_close = _NodeLocalQueuesMethods._process__locals_queue_close
	_process__locals_queues_len = _NodeLocalQueuesMethods._process__locals_queues_len
	_process__locals_queues_contains = _NodeLocalQueuesMethods._process__locals_queues_contains
	_process__locals_queues_keys = _NodeLocalQueuesMethods._process__locals_queues_keys
	_process__locals_queues_iter = _NodeLocalQueuesMethods._process__locals_queues_iter
	_process__locals_queues_delitem = _NodeLocalQueuesMethods._process__locals_queues_delitem
	_process__locals_queues_clear = _NodeLocalQueuesMethods._process__locals_queues_clear

	# local shared dict
	__getitem__ = _NodeLocalDictMethods.__getitem__
	__setitem__ = _NodeLocalDictMethods.__setitem__
	__delitem__ = _NodeLocalDictMethods.__delitem__
	__iter__ = _NodeLocalDictMethods.__iter__
	__contains__ = _NodeLocalDictMethods.__contains__
	__len__ = _NodeLocalDictMethods.__len__
	clear = _NodeLocalDictMethods.clear
	keys = _NodeLocalDictMethods.keys
	values = _NodeLocalDictMethods.values
	items = _NodeLocalDictMethods.items
	pop = _NodeLocalDictMethods.pop
	_process___getitem__ = _NodeLocalDictMethods._process___getitem__
	_process___setitem__ = _NodeLocalDictMethods._process___setitem__
	_process___delitem__ = _NodeLocalDictMethods._process___delitem__
	_process___iter__ = _NodeLocalDictMethods._process___iter__
	_process___contains__ = _NodeLocalDictMethods._process___contains__
	_process___len__ = _NodeLocalDictMethods._process___len__
	_process_clear = _NodeLocalDictMethods._process_clear
	_process_keys = _NodeLocalDictMethods._process_keys
	_process_values = _NodeLocalDictMethods._process_values
	_process_items = _NodeLocalDictMethods._process_items
	_process_pop = _NodeLocalDictMethods._process_pop

	# global queues
	_globals_get = _NodeGlobalQueuesMethods._globals_get
	_globals_put = _NodeGlobalQueuesMethods._globals_put
	_globals_qsize = _NodeGlobalQueuesMethods._globals_qsize
	_globals_queue_get = _NodeGlobalQueuesMethods._globals_queue_get
	_globals_queue_put = _NodeGlobalQueuesMethods._globals_queue_put
	_globals_queue_close = _NodeGlobalQueuesMethods._globals_queue_close
	_globals_queue_len = _NodeGlobalQueuesMethods._globals_queue_len
	_globals_queues_len = _NodeGlobalQueuesMethods._globals_queues_len
	_globals_queues_delitem = _NodeGlobalQueuesMethods._globals_queues_delitem
	_globals_queues_contains = _NodeGlobalQueuesMethods._globals_queues_contains
	_globals_queues_keys = _NodeGlobalQueuesMethods._globals_queues_keys
	_globals_queues_iter = _NodeGlobalQueuesMethods._globals_queues_iter
	_globals_queues_clear = _NodeGlobalQueuesMethods._globals_queues_clear
	_process__globals_get = _NodeGlobalQueuesMethods._process__globals_get
	_process__globals_put = _NodeGlobalQueuesMethods._process__globals_put
	_process__globals_qsize = _NodeGlobalQueuesMethods._process__globals_qsize
	_process__globals_queue_get = _NodeGlobalQueuesMethods._process__globals_queue_get
	_process__globals_queue_put = _NodeGlobalQueuesMethods._process__globals_queue_put
	_process__globals_queue_close = _NodeGlobalQueuesMethods._process__globals_queue_close
	_process__globals_queue_len = _NodeGlobalQueuesMethods._process__globals_queue_len
	_process__globals_queues_len = _NodeGlobalQueuesMethods._process__globals_queues_len
	_process__globals_queues_delitem = _NodeGlobalQueuesMethods._process__globals_queues_delitem
	_process__globals_queues_contains = _NodeGlobalQueuesMethods._process__globals_queues_contains
	_process__globals_queues_keys = _NodeGlobalQueuesMethods._process__globals_queues_keys
	_process__globals_queues_iter = _NodeGlobalQueuesMethods._process__globals_queues_iter
	_process__globals_queues_clear = _NodeGlobalQueuesMethods._process__globals_queues_clear

	# global shared dict
	_globals_getitem = _NodeGlobalDictMethods._globals_getitem
	_globals_setitem = _NodeGlobalDictMethods._globals_setitem
	_globals_delitem = _NodeGlobalDictMethods._globals_delitem
	_globals_pop = _NodeGlobalDictMethods._globals_pop
	_globals_iter = _NodeGlobalDictMethods._globals_iter
	_globals_contains = _NodeGlobalDictMethods._globals_contains
	_globals_len = _NodeGlobalDictMethods._globals_len
	_globals_keys = _NodeGlobalDictMethods._globals_keys
	_globals_values = _NodeGlobalDictMethods._globals_values
	_globals_items = _NodeGlobalDictMethods._globals_items
	_globals_clear = _NodeGlobalDictMethods._globals_clear
	_process__globals_getitem = _NodeGlobalDictMethods._process__globals_getitem
	_process__globals_setitem = _NodeGlobalDictMethods._process__globals_setitem
	_process__globals_delitem = _NodeGlobalDictMethods._process__globals_delitem
	_process__globals_clear = _NodeGlobalDictMethods._process__globals_clear
	_process__globals_keys = _NodeGlobalDictMethods._process__globals_keys
	_process__globals_values = _NodeGlobalDictMethods._process__globals_values
	_process__globals_items = _NodeGlobalDictMethods._process__globals_items
	_process__globals_iter = _NodeGlobalDictMethods._process__globals_iter
	_process__globals_contains = _NodeGlobalDictMethods._process__globals_contains
	_process__globals_len = _NodeGlobalDictMethods._process__globals_len
	_process__globals_pop = _NodeGlobalDictMethods._process__globals_pop

	# file transfer
	get_file = _NodeFileTransferMethods.get_file
	_process_get_file = _NodeFileTransferMethods._process_get_file
	put_file = _NodeFileTransferMethods.put_file
	_process_put_file = _NodeFileTransferMethods._process_put_file
	get_folder = _NodeFileTransferMethods.get_folder
	_process_get_folder = _NodeFileTransferMethods._process_get_folder
	put_folder = _NodeFileTransferMethods.put_folder
	_process_put_folder = _NodeFileTransferMethods._process_put_folder
	_write_to_file = _NodeFileTransferMethods._write_to_file
	_process__write_to_file = _NodeFileTransferMethods._process__write_to_file
	_close_file = _NodeFileTransferMethods._close_file
	_process__close_file = _NodeFileTransferMethods._process__close_file
	
	# RPC
	eval = _NodeRPCMethods.eval
	exec = _NodeRPCMethods.exec
	execfile = _NodeRPCMethods.execfile
	exec_remote_file = _NodeRPCMethods.exec_remote_file
	system = _NodeRPCMethods.system
	call = _NodeRPCMethods.call
	call_remote = _NodeRPCMethods.call_remote
	_process_system = _NodeRPCMethods._process_system
	_process_eval = _NodeRPCMethods._process_eval
	_process_exec = _NodeRPCMethods._process_exec
	_process_exec_remote_file = _NodeRPCMethods._process_exec_remote_file
	_process_call = _NodeRPCMethods._process_call
	_process_call_remote = _NodeRPCMethods._process_call_remote

	@property
	def is_closed(self):
		try:
			return self._connection._closed
		except:
			return True

	@property
	def address(self):
		return self._address

	@property
	def server_address(self):
		return self._server_address

	@property
	def server(self):
		return self._server

	@property
	def locals(self):
		return self

	@property
	def globals(self):
		return self._globals

	@property
	def queues(self):
		return self._queues

	@property
	def pipes(self):
		return self._pipes


import threading
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _utils import file_size, md5, eprint
from _NodeInternalClasses import Future, Thread

def _write_to_file(self, data, filename):
	self._request(-2, data=data, file_name=filename)

def _process__write_to_file(self, request):
	try:
		filename = os.path.abspath(request["data"]["file_name"])
		if self._out_file != None and self._out_file.name != filename:
			self._out_file.close()
			self._out_file = None

		if self._out_file == None:
			if not os.path.isdir(os.path.dirname(filename)):
				os.makedirs(os.path.dirname(filename))

			self._out_file = open(filename, "wb")

		self._out_file.write(request["data"]["data"])
		self._out_file.flush()
	except:
		pass

def _close_file(self):
	self._request(-2)

def _process__close_file(self, request):
	try:
		if self._out_file != None:
			self._out_file.close()
			self._out_file = None
	except:
		pass

def get_file(self, src_filename, dest_filename = None, block=True):
	session_id = self._get_session_id()

	if dest_filename == None:
		dest_filename = os.path.basename(src_filename)

	self._request(session_id, src_filename=src_filename, file_size=file_size(dest_filename), md5=md5(dest_filename), block=block)
	def session():
		response = self._recv_response(session_id)
		if response["cancel"] or response["data"]["same_file"]:
			return

		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]
		
		if not os.path.isdir(os.path.dirname(os.path.abspath(dest_filename))):
			os.makedirs(os.path.dirname(os.path.abspath(dest_filename)))

		try:
			file = open(dest_filename, "wb")
			self._respond_ok(session_id)
			file.close()
		except BaseException as e:
			self._respond_exception(session_id, e)
			if block:
				raise e
			else:
				return

		recved_size = 0
		full_size = response["data"]["file_size"]
		with open(dest_filename, "wb") as file:
			while recved_size < full_size:
				response = self._recv_response(session_id)
				if response["cancel"]:
					break
				file.write(response["data"]["data"])
				file.flush()
				recved_size += len(response["data"]["data"])

	if block:
		session()
	else:
		thread = threading.Thread(target=session)
		thread.start()
		return Future(self, session_id)

def _process_get_file(self, request):
	session_id = request["session_id"]

	def session():
		src_filename = request["data"]["src_filename"]
		src_file_size = file_size(src_filename)

		same_file = (src_file_size != 0 and request["data"]["file_size"] == src_file_size and \
			         request["data"]["md5"] == md5(src_filename))
		try:
			file = open(src_filename, "rb")
			file.close()
			self._respond_ok(session_id, same_file=same_file, file_size=src_file_size)
		except BaseException as e:
			self._respond_exception(session_id, e, same_file=same_file, block=request["block"])
			self._make_signal(session_id)
			return

		if same_file:
			if not request["block"]:
				self._put_result(session_id)
			self._make_signal(session_id)
			return

		response = self._recv_response(session_id)
		if not response["success"]:
			if not request["block"]:
				self._put_exception(session_id, response["exception"])
			self._make_signal(session_id)
			return
		
		block_size = 8192*1024
		sent_size = 0
		with open(src_filename, "rb") as file:
			while sent_size < src_file_size:
				data = file.read(block_size)
				self._respond_ok(session_id, data=data)
				sent_size += len(data)

		if not request["block"]:
			self._put_result(session_id)

		self._make_signal(session_id)

	thread = Thread(target=session)
	thread.start()
	signal = self._recv_signal(session_id)
	if signal["cancel"] and thread.is_alive():
		thread.kill()
		thread.join()
		self._respond_ok(session_id, cancel=True)
		self._put_result(session_id, cancelled=True)
	else:
		thread.join()

def put_file(self, src_filename, dest_filename = None, block=True):
	file = open(src_filename, "rb")
	file.close()

	session_id = self._get_session_id()

	if dest_filename == None:
		dest_filename = os.path.basename(src_filename)

	src_file_size = file_size(src_filename)

	self._stop_send = False
	self._request(session_id, md5=md5(src_filename), file_size=src_file_size, dest_filename=dest_filename, block=block)
	
	def session():
		response = self._recv_response(session_id)
		if response["cancel"] or response["data"]["same_file"]:
			return

		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

		block_size = 8192*1024
		sent_size = 0
		with open(src_filename, "rb") as file:
			while sent_size < src_file_size:
				if self._stop_send:
					break
				data = file.read(block_size)
				self._respond_ok(session_id, data=data)
				sent_size += len(data)

	if block:
		session()
	else:
		thread = Thread(target=session)
		thread.start()
		return Future(self, session_id)

def _process_put_file(self, request):
	session_id = request["session_id"]
	def session():
		dest_filename = request["data"]["dest_filename"]
		full_size = request["data"]["file_size"]

		if full_size == file_size(dest_filename) and \
		   request["data"]["md5"] == md5(dest_filename):
			self._respond_ok(session_id, same_file=True, block=request["block"])
			self._make_signal(session_id)
			return
		
		if not os.path.isdir(os.path.dirname(os.path.abspath(dest_filename))):
			os.makedirs(os.path.dirname(os.path.abspath(dest_filename)))
		
		try:
			file = open(dest_filename, "wb")
			file.close()
			self._respond_ok(session_id, same_file=False)
		except BaseException as e:
			self._respond_exception(session_id, e, same_file=False, block=request["block"])
			self._make_signal(session_id)
			return

		recved_size = 0
		with open(dest_filename, "wb") as file:
			while recved_size < full_size:
				response = self._recv_response(session_id)					
				file.write(response["data"]["data"])
				file.flush()
				recved_size += len(response["data"]["data"])

		if not request["block"]:
			self._put_result(session_id)

		self._make_signal(session_id)

	thread = Thread(target=session)
	thread.start()
	signal = self._recv_signal(session_id)
	if signal["cancel"] and thread.is_alive():
		thread.kill()
		thread.join()
		self._respond_ok(session_id, cancel=True)
		self._put_result(session_id, cancelled=True)
	else:
		thread.join()

def get_folder(self, src_foldername, dest_foldername = None, block=True):
	session_id = self._get_session_id()
	if dest_foldername == None:
		dest_foldername = os.path.basename(src_foldername)

	self._request(session_id, src_foldername=src_foldername, block=block)
	# self._request(session_id, src_foldername=src_foldername, block=block, debug="I need folder \"" + src_foldername + "\"")
	def session():
		response = self._recv_response(session_id)
		if response["cancel"]:
			return

		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

		for folder in response["data"]["folders"]:
			if not os.path.isdir(dest_foldername + "/" + folder):
				try:
					os.makedirs(dest_foldername + "/" + folder)
				except:
					eprint(self._traceback(False))
		self._respond_ok(session_id)
		# self._respond_ok(session_id, debug="Created all folders. Please send files.")

		while True:
			response = self._recv_response(session_id)
			if response["cancel"] or response["data"]["finished"]:
				return

			if not response["success"]:
				eprint(response["traceback"])
				continue

			dest_filename = dest_foldername + "/" + response["data"]["file_name"]
			same_file = (response["data"]["file_size"] == file_size(dest_filename) and \
				         response["data"]["md5"] == md5(dest_filename))

			if same_file:
				self._respond_ok(session_id, same_file = same_file)
				# self._respond_ok(session_id, same_file = same_file, debug="Already have file: " + dest_filename)
				continue
			else:
				try:
					file = open(dest_filename, "wb")
					self._respond_ok(session_id, same_file = same_file)
					# self._respond_ok(session_id, same_file = same_file, debug="I don't have this file, please send me.")
					file.close()
				except BaseException as e:
					eprint(self._traceback())
					self._respond_exception(session_id, e)
					# self._respond_exception(session_id, e, debug="Error")
					continue

			recved_size = 0
			full_size = response["data"]["file_size"]
			with open(dest_filename, "wb") as file:
				while recved_size < full_size:
					response = self._recv_response(session_id)
					if response["cancel"]:
						return
					file.write(response["data"]["data"])
					file.flush()
					recved_size += len(response["data"]["data"])

	if block:
		session()
	else:
		thread = threading.Thread(target=session)
		thread.start()
		return Future(self, session_id)

def _process_get_folder(self, request):
	session_id = request["session_id"]
	def session():
		src_foldername = request["data"]["src_foldername"]
		if not os.path.isdir(src_foldername):
			self._respond_exception(session_id, FileNotFoundError(src_foldername + " is not a folder."), block=request["block"])
			self._make_signal(session_id)
			return
		i = len(src_foldername)-1
		while src_foldername[i] in ['/', '\\']:
			i -= 1

		folders = []
		for root, dirs, files in os.walk(src_foldername):
			for name in dirs:
				folders.append(os.path.join(root, name)[i+2:])
		self._respond_ok(session_id, folders=folders)
		# self._respond_ok(session_id, folders=folders, debug="Please create folders: " + str(folders))
		response = self._recv_response(session_id)

		for root, dirs, files in os.walk(src_foldername):
			for name in files:
				src_filename = os.path.join(root, name)
				src_file_size = file_size(src_filename)
				relative_file_name = src_filename[i+2:]
				
				try:
					file = open(src_filename, "rb")
					file.close()
					self._respond_ok(session_id, finished=False, file_name=relative_file_name, file_size=src_file_size, md5=md5(src_filename))
					# self._respond_ok(session_id, finished=False, file_name=relative_file_name, file_size=src_file_size, md5=md5(src_filename), debug="Do you have file " + src_filename + "?")
				except BaseException as e:
					self._respond_exception(session_id, e)
					# self._respond_exception(session_id, e, debug="I cannot open file: " + src_filename)
					continue

				response = self._recv_response(session_id)
				if not response["success"] or response["data"]["same_file"]:
					continue

				block_size = 8192*1024
				sent_size = 0
				with open(src_filename, "rb") as file:
					while sent_size < src_file_size:
						data = file.read(block_size)
						self._respond_ok(session_id, data=data)
						# self._respond_ok(session_id, data=data, debug="Please write " + str(len(data)) + " bytes data to file " + relative_file_name)
						sent_size += len(data)

		self._respond_ok(session_id, finished=True)
		# self._respond_ok(session_id, finished=True, debug="Sent all files.")

		if not request["block"]:
			self._put_result(session_id)

		self._make_signal(session_id)

	thread = Thread(target=session)
	thread.start()
	signal = self._recv_signal(session_id)
	if signal["cancel"] and thread.is_alive():
		thread.kill()
		thread.join()
		self._respond_ok(session_id, cancel=True)
		# self._respond_ok(session_id, cancel=True, debug="Cancel.")
		self._put_result(session_id, cancelled=True)
	else:
		thread.join()

def put_folder(self, src_foldername, dest_foldername = None, block = True):
	if not os.path.isdir(src_foldername):
		raise FileNotFoundError(src_foldername + " is not a folder.")

	session_id = self._get_session_id()

	if dest_foldername == None:
		dest_foldername = os.path.basename(src_foldername)

	self._stop_send = False
	self._request(session_id, dest_foldername=dest_foldername, block=block)
	# self._request(session_id, dest_foldername=dest_foldername, block=block, debug="I will send you folder: " + dest_foldername)
	def session():
		response = self._recv_response(session_id)
		if response["cancel"]:
			return

		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

		i = len(src_foldername)-1
		while src_foldername[i] in ['/', '\\']:
			i -= 1

		folders = []
		for root, dirs, files in os.walk(src_foldername):
			for name in dirs:
				folders.append(os.path.join(root, name)[i+2:])
		self._respond_ok(session_id, folders=folders)
		# self._respond_ok(session_id, folders=folders, debug="Please create folders: " + str(folders))
		response = self._recv_response(session_id)
		if response["cancel"]:
			return

		for root, dirs, files in os.walk(src_foldername):
			for name in files:
				src_filename = os.path.join(root, name)
				src_file_size = file_size(src_filename)
				relative_file_name = src_filename[i+2:]
				
				try:
					file = open(src_filename, "rb")
					file.close()
					self._respond_ok(session_id, finished=False, file_name=relative_file_name, file_size=src_file_size, md5=md5(src_filename))
					# self._respond_ok(session_id, finished=False, file_name=relative_file_name, file_size=src_file_size, md5=md5(src_filename), debug="Do you have file " + relative_file_name + "?")
				except BaseException as e:
					self._respond_exception(session_id, e) # , debug="Error")
					continue

				response = self._recv_response(session_id)
				if response["cancel"]:
					return

				if not response["success"] or response["data"]["same_file"]:
					continue

				block_size = 8192*1024
				sent_size = 0
				with open(src_filename, "rb") as file:
					while sent_size < src_file_size:
						if self._stop_send:
							return
						data = file.read(block_size)
						self._respond_ok(session_id, data=data)
						# self._respond_ok(session_id, data=data, debug="Please write " + str(len(data)) + " bytes to file " + relative_file_name)
						sent_size += len(data)

		self._respond_ok(session_id, finished=True)
		# self._respond_ok(session_id, finished=True, debug="Sent all files.")

	if block:
		session()
	else:
		thread = threading.Thread(target=session)
		thread.start()
		return Future(self, session_id)

def _process_put_folder(self, request):
	session_id = request["session_id"]

	def session():
		dest_foldername = request["data"]["dest_foldername"]
		try:
			if not os.path.isdir(dest_foldername):
				os.makedirs(dest_foldername)
			self._respond_ok(session_id)
			# self._respond_ok(session_id, debug="Created folder " + dest_foldername)
		except BaseException as e:
			self._respond_exception(session_id, e, block=request["block"])
			# self._respond_exception(session_id, e, block=request["block"], debug="Cannot create folder " + dest_foldername)
			self._make_signal(session_id)
			return

		response = self._recv_response(session_id)
		for folder in response["data"]["folders"]:
			if not os.path.isdir(dest_foldername + "/" + folder):
				try:
					os.makedirs(dest_foldername + "/" + folder)
				except:
					pass
		self._respond_ok(session_id)
		# self._respond_ok(session_id, debug="Created all folders.")

		while True:
			response = self._recv_response(session_id)
			if not response["success"]:
				continue

			if response["data"]["finished"]:
				break

			dest_filename = dest_foldername + "/" + response["data"]["file_name"]
			same_file = (response["data"]["file_size"] == file_size(dest_filename) and \
				         response["data"]["md5"] == md5(dest_filename))

			if same_file:
				self._respond_ok(session_id, same_file = same_file)
				# self._respond_ok(session_id, same_file = same_file, debug="Already have file " + dest_filename)
				continue
			else:
				try:
					file = open(dest_filename, "wb")
					self._respond_ok(session_id, same_file = same_file)
					# self._respond_ok(session_id, same_file = same_file, debug="I don't have this file, please send me.")
					file.close()
				except BaseException as e:
					self._respond_exception(session_id, e)
					# self._respond_exception(session_id, e, debug="Error")
					continue

			recved_size = 0
			full_size = response["data"]["file_size"]
			with open(dest_filename, "wb") as file:
				while recved_size < full_size:
					response = self._recv_response(session_id)
					file.write(response["data"]["data"])
					file.flush()
					recved_size += len(response["data"]["data"])
		
		if not request["block"]:
			self._put_result(session_id)

		self._make_signal(session_id)

	thread = Thread(target=session)
	thread.start()
	signal = self._recv_signal(session_id)
	if signal["cancel"] and thread.is_alive():
		thread.kill()
		thread.join()
		self._respond_ok(session_id, cancel=True)
		# self._respond_ok(session_id, cancel=True, debug="Cancel.")
		self._put_result(session_id, cancelled=True)
	else:
		thread.join()	


import os
import sys
import subprocess

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _NodeInternalClasses import Future, Thread, AutoSendBuffer
from _utils import eprint

def eval(self, statement, block=True):
	session_id = self._get_session_id()

	self._request(session_id, statement=statement, block=block)
	if block:
		response = self._recv_response(session_id)

		if response["success"]:
			return response["data"]["return_value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]
	else:
		return Future(self, session_id)

def exec(self, code, block=True):
	session_id = self._get_session_id()

	self._request(session_id, code=code, block=block)
	if block:
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]
	else:
		return Future(self, session_id)

def system(self, cmd, quiet=False, remote_quiet=False, once_all=False, block=True):
	session_id = self._get_session_id()

	self._request(session_id, cmd=cmd, remote_quiet=remote_quiet, once_all=once_all, block=block)
	if not once_all:
		def session(session_id, quiet):
			while True:
				response = self._recv_response(session_id)
				if response["success"]:
					if not response["data"]["end"]:
						if not quiet:
							print(response["data"]["stdout"], end="", flush=True)
					else:
						return response["data"]["return_value"]
				else:
					eprint(response["traceback"])
					raise response["exception"]

		if block:
			session(session_id, quiet)
		else:
			thread = Thread(target=session, args=(session_id, quiet))
			thread.start()
			return Future(self, session_id)
	else:
		if block:
			response = self._recv_response(session_id)
			if response["success"]:
				if not quiet:
					print(response["data"]["stdout"], end="", flush=True)
					return response["data"]["return_value"]
			else:
				eprint(response["traceback"])
				raise response["exception"]
		else:
			return Future(self, session_id)

# call local function on remote computer
def call(self, target, args=(), kwargs={}, block=True):
	session_id = self._get_session_id()

	self._request(session_id, target=target, args=args, kwargs=kwargs, block=block)
	if block:
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["return_value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]
	else:
		return Future(self, session_id)

# call remote function on remote computer
def call_remote(self, target, args=(), kwargs={}, block=True):
	session_id = self._get_session_id()

	self._request(session_id, target=target, args=args, kwargs=kwargs, block=block)
	if block:
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["return_value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]
	else:
		return Future(self, session_id)

def execfile(self, script_name, block=True):
	self.put_file(script_name, ".temp_scripts/" + os.path.basename(script_name))
	return self.exec_remote_file(".temp_scripts/" + os.path.basename(script_name), block=block)

def exec_remote_file(self, script_name, block=True):
	session_id = self._get_session_id()

	self._request(session_id, script_name=script_name, block=block)
	if block:
		response = self._recv_response(session_id)

		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]
	else:
		return Future(self, session_id)

def _process_system(self, request):
	session_id = request["session_id"]

	try:
		if not request["data"]["once_all"]:
			process = subprocess.Popen(request["data"]["cmd"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
			self.__sent_end = False
			self.__did_put_result = False
			def session():
				remote_quiet = request["data"]["remote_quiet"]
				auto_send_buffer = AutoSendBuffer(self, session_id, timeout=0.1, quiet=remote_quiet)
				while True:
					byte = process.stdout.read(1)
					if byte == b'':
						break
					auto_send_buffer.append(byte)
				auto_send_buffer.stop()
				process.communicate()
				self._respond_ok(session_id, end=True, return_value=process.returncode)
				self.__sent_end = True
				if not request["block"]:
					self._put_result(session_id, return_value=process.returncode)
					self.__did_put_result = True
				self._make_signal(session_id)

			thread = Thread(target=session)
			thread.start()
			signal = self._recv_signal(session_id)
			if signal["cancel"] and thread.is_alive():
				process.kill()
				thread.kill()
				thread.join()
				if not self.__sent_end:
					self._respond_ok(session_id, end=True, return_value=0, block=request["block"])
				if not self.__did_put_result and not request["block"]:
					self._put_result(session_id, cancel=True)
			else:
				thread.join()
		else:
			process = subprocess.Popen(request["data"]["cmd"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
			def session():
				stdout, stderr = process.communicate()

				stdout = stdout.decode("utf-8")
				if not request["data"]["remote_quiet"]:
					print(stdout, end="", flush=True)

				self._respond_ok(session_id, stdout=stdout, return_value=process.returncode, block=request["block"])
				self._make_signal(session_id)

			thread = Thread(target=session)
			thread.start()
			signal = self._recv_signal(session_id)
			if signal["cancel"] and thread.is_alive():
				process.kill()
				thread.kill()
				thread.join()
				self._put_result(session_id, cancelled=True)
			else:
				thread.join()
			
	except BaseException as e:
		self._respond_exception(session_id, e, block=(request["block"] or not request["data"]["once_all"]))
		
def _process_eval(self, request):
	session_id = request["session_id"]

	try:
		def session():
			try:
				return_value = eval(request["data"]["statement"])
				self._respond_ok(session_id, return_value=return_value, block=request["block"])
			except BaseException as e:
				self._respond_exception(session_id, e, block=request["block"])
			self._make_signal(session_id)

		thread = Thread(target=session)
		thread.start()
		signal = self._recv_signal(session_id)
		if signal["cancel"] and thread.is_alive():
			thread.kill()
			thread.join()
			self._put_result(session_id, cancelled=True)
		else:
			thread.join()
		
	except BaseException as e:
		self._respond_exception(session_id, e, block=request["block"])
		
def _process_exec(self, request):
	session_id = request["session_id"]

	try:
		def session():
			try:
				exec(request["data"]["code"], globals())
				self._respond_ok(session_id, block=request["block"])
			except BaseException as e:
				self._respond_exception(session_id, e, block=request["block"])

			self._make_signal(session_id)

		thread = Thread(target=session)
		thread.start()
		signal = self._recv_signal(session_id)
		if signal["cancel"] and thread.is_alive():
			thread.kill()
			thread.join()
			self._put_result(session_id, cancelled=True)
		else:
			thread.join()
	except BaseException as e:
		self._respond_exception(session_id, e, block=request["block"])

def _process_call(self, request):
	session_id = request["session_id"]

	try:
		def session():
			try:
				return_value = request["data"]["target"](*request["data"]["args"], **request["data"]["kwargs"])
				self._respond_ok(session_id, return_value=return_value, block=request["block"])
			except BaseException as e:
				self._respond_exception(session_id, e, block=request["block"])

			self._make_signal(session_id)

		thread = Thread(target=session)
		thread.start()
		signal = self._recv_signal(session_id)
		if signal["cancel"] and thread.is_alive():
			thread.kill()
			thread.join()
			self._put_result(session_id, cancelled=True)
		else:
			thread.join()
	except BaseException as e:
		self._respond_exception(session_id, e, block=request["block"])

def _process_call_remote(self, request):
	session_id = request["session_id"]

	try:
		def session():
			try:
				return_value = eval(request["data"]["target"])(*request["data"]["args"], **request["data"]["kwargs"])
				self._respond_ok(session_id, return_value=return_value, block=request["block"])
			except BaseException as e:
				self._respond_exception(session_id, e, block=request["block"])

			self._make_signal(session_id)

		thread = Thread(target=session)
		thread.start()
		signal = self._recv_signal(session_id)
		if signal["cancel"] and thread.is_alive():
			thread.kill()
			thread.join()
			self._put_result(session_id, cancelled=True)
		else:
			thread.join()

	except BaseException as e:
		self._respond_exception(session_id, e, block=request["block"])

def _process_exec_remote_file(self, request):
	session_id = request["session_id"]

	try:
		def session():
			try:
				exec(open(request["data"]["script_name"]).read(), globals())
				self._respond_ok(session_id, block=request["block"])
			except BaseException as e:
				self._respond_exception(session_id, e, block=request["block"])
			
			self._make_signal(session_id)

		thread = Thread(target=session)
		thread.start()
		signal = self._recv_signal(session_id)
		if signal["cancel"] and thread.is_alive():
			thread.kill()
			thread.join()
			self._put_result(session_id, cancelled=True)
		else:
			thread.join()

	except BaseException as e:
		self._respond_exception(session_id, e, block=request["block"])


import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _utils import eprint

def _globals_getitem(self, name):
	if self._server != None:
		return self._server[name]
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_setitem(self, name, value):
	if self._server != None:
		self._server[name] = value
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, value=value)
		response = self._recv_response(session_id)
		
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_delitem(self, name):
	if self._server != None:
		del self._server[name]
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_pop(self, name):
	if self._server != None:
		return self._server.pop(name)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_iter(self):
	if self._server != None:
		return self._server.__iter__()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["iter"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_contains(self, name):
	if self._server != None:
		return self._server.__contains__(name)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["contains"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_len(self):
	if self._server != None:
		return self._server.__len__()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["len"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_keys(self):
	if self._server != None:
		return self._server.keys()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["keys"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_values(self):
	if self._server != None:
		return self._server.values()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["values"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_items(self):
	if self._server != None:
		return self._server.items()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["items"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _globals_clear(self):
	if self._server != None:
		self._server.clear()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_getitem(self, request):
	try:
		self._respond_ok(request["session_id"], value=self._server[request["data"]["name"]])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_setitem(self, request):
	try:
		self._server[request["data"]["name"]] = request["data"]["value"]
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_delitem(self, request):
	try:
		del self._server[request["data"]["name"]]
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_clear(self, request):
	try:
		self._server.clear()
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_keys(self, request):
	try:
		self._respond_ok(request["session_id"], keys=self._server.keys())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_values(self, request):
	try:
		self._respond_ok(request["session_id"], values=self._server.values())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_items(self, request):
	try:
		self._respond_ok(request["session_id"], items=self._server.items())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_iter(self, request):
	try:
		self._respond_ok(request["session_id"], iter=self._server.__iter__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_contains(self, request):
	try:
		self._respond_ok(request["session_id"], contains=self._server.__contains__(request["data"]["name"]))
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_len(self, request):
	try:
		self._respond_ok(request["session_id"], len=self._server.__len__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process__globals_pop(self, request):
	try:
		self._respond_ok(request["session_id"], value=self._server.pop(request["data"]["name"]))
	except BaseException as e:
		self._respond_exception(request["session_id"], e)


import threading
import cloudpickle
import pickle
import multiprocessing
import sys
import os
import traceback
from concurrent.futures import ThreadPoolExecutor

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
import _NodeInternalClasses

complex_request = ["get_file", "put_file",
                   "get_folder", "put_folder",
	               "eval", "exec", "system",
	               "call", "call_remote",
	               "execfile", "exec_remote_file"]

def __init__(self, connection, server = None):
	# session id:
	# request a new session id use -1
	# close, write_to_file, close_file use -2

	self._is_Node_init = True
	self._connection = connection

	self._server = server

	self._pipes = _NodeInternalClasses.Pipes(self)
	self._queues = _NodeInternalClasses.Queues(self)
	self._result_queues_for_future = _NodeInternalClasses.Queues(self, private=True)

	if self._server == None:
		self._local_vars = {}
		self._pipe = multiprocessing.Pipe()
		self._queue = _NodeInternalClasses.CloseableQueue()
		self._actual_pipes = {}
		self._actual_queues = {}
		self._actual_result_queues_for_future = {}
		self._address = self._connection.getsockname()
		self._server_address = self._connection.getpeername()
	else:
		self._next_session_id = 0
		self._address = self._connection.getpeername()
		self._server_address = server.address
		
	self._globals = _NodeInternalClasses.Globals(self)
	self._recving_thread = None

	self._recved_responses = _NodeInternalClasses.QueueDict()
	self._recved_signals = _NodeInternalClasses.QueueDict()
	
	self._simple_threads_pool = None
	self._complex_threads_pool = None

	self._out_file = None

	self._start()

def __del__(self):
	try:
		self.close()
	except:
		pass

def hold_on(self):
	threading.Event().wait()

def close(self):
	if self._server == None:
		try:
			self._request(session_id=-2)
		except:
			pass

		try:
			self._connection.close()
		except:
			pass

		self._connection = None
		self._address = None
		self._server_address = None

		try:
			self._queue.close()
		except:
			pass

		try:
			self._pipe[0].close()
		except:
			pass

		try:
			self._pipe[1].close()
		except:
			pass

		for key in self._actual_queues:
			try:
				self._actual_queues[key].close()
			except:
				pass
		self._actual_queues.clear()

		for key in self._actual_pipes:
			try:
				self._actual_pipes[key][0].close()
			except:
				pass

			try:
				self._actual_pipes[key][1].close()
			except:
				pass
		self._actual_pipes.clear()

		for key in self._actual_result_queues_for_future:
			try:
				self._actual_result_queues_for_future[key].close()
			except:
				pass
		self._actual_result_queues_for_future.clear()

	else:
		try:
			self._connection.close()
		except:
			pass

		self._connection = None
		
		try:
			if self.address != None:
				self._server.disconnect_result[self.address] = self._server._callback_threads_pool.submit(self._server._onDisconnect, self.address)
		except:
			pass

		try:
			self._server._connected_nodes.pop(self.address)
		except:
			pass

		self._address = None

	if self._simple_threads_pool != None:
		try:
			self._simple_threads_pool.shutdown()
		except:
			pass
		self._simple_threads_pool = None

	if self._complex_threads_pool != None:
		try:
			self._complex_threads_pool.shutdown()
		except:
			pass
		self._complex_threads_pool = None

	try:
		self._recved_signals.clear()
	except:
		pass

	try:
		self._recved_responses.clear()
	except:
		pass

def _start(self):
	self._simple_threads_pool = ThreadPoolExecutor(max_workers=5)
	self._complex_threads_pool = ThreadPoolExecutor(max_workers=1)
	if self._recving_thread == None or not self._recving_thread.is_alive():
		self._recving_thread = threading.Thread(target=self._recving_loop, daemon=True)
		self._recving_thread.start()

def _get_session_id(self):
	if self._server != None:
		session_id = self._next_session_id
		self._next_session_id += 1
		if self._next_session_id >= 1024:
			self._next_session_id = 0
		return session_id
	else:
		self._request(session_id=-1)
		response = self._recv_response(session_id=-1)
		return response["data"]["target_session_id"]

def _traceback(self, remote=True):
	exception_list = traceback.format_stack()
	exception_list = exception_list[:-2]
	exception_list.extend(traceback.format_tb(sys.exc_info()[2]))
	exception_list.extend(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1]))

	exception_str = "Traceback (most recent call last):\n"
	if remote:
		exception_str = "Remote"+str(self._server_address) + " " + exception_str

	exception_str += "".join(exception_list)

	return exception_str

def _process__get_session_id(self, request):
	self._respond_ok(session_id=-1, target_session_id=self._next_session_id)
	self._next_session_id += 1
	if self._next_session_id >= 1024:
		self._next_session_id = 0

def _process_close(self, request):
	self.close()

def _send(self, value):
	binary = cloudpickle.dumps(value, 3)
	length = len(binary)
	self._connection.sendall(length.to_bytes(4, byteorder='little', signed=False) + binary)

def _send_signal(self, session_id, **kwargs):
	cancel = False
	if "cancel" in kwargs:
		cancel = kwargs.pop("cancel")

	signal = \
	{
		"session_id": session_id,
		"type": "signal",
		"cancel": cancel,
		"data": kwargs
	}
	self._send(signal)

def _request(self, session_id, **kwargs):
	block = True
	if "block" in kwargs:
		block = kwargs.pop("block")

	if "debug" in kwargs:
		print("request:", kwargs["debug"], flush=True)

	func = sys._getframe().f_back.f_code.co_name
	if "func" in kwargs:
		func = kwargs.pop("func")

	request = \
	{
		"session_id": session_id,
		"type": "request",
		"func": func,
		"block": block,
		"data": kwargs
	}
	self._send(request)

def _respond_ok(self, session_id, **kwargs):
	block = True
	if "block" in kwargs:
		block = kwargs.pop("block")

	cancel = False
	if "cancel" in kwargs:
		cancel = kwargs.pop("cancel")

	if "debug" in kwargs:
		print("respond:", kwargs["debug"], flush=True)

	if block:
		response = \
		{
			"session_id": session_id,
			"type": "response",
			"success": (not cancel),
			"cancel": cancel,
			"data": kwargs
		}
		self._send(response)
	else:
		self._put_result(session_id, **kwargs)

def _respond_exception(self, session_id, exception, **kwargs):
	block = True
	if "block" in kwargs:
		block = kwargs.pop("block")

	if "debug" in kwargs:
		print("respond exception:", kwargs["debug"], flush=True)

	if block:
		response = \
		{
			"session_id": session_id,
			"type": "response",
			"success": False,
			"exception": exception,
			"traceback": self._traceback(),
			"data": kwargs
		}
		self._send(response)
	else:
		self._put_exception(session_id, exception, **kwargs)

def _put_result(self, session_id, **kwargs):
	result = dict(success=True, return_value=None, exception=None, traceback=None, cancelled=False)
	for key in result:
		if key in kwargs:
			result[key] = kwargs.pop(key)
	result["data"] = kwargs

	self._result_queues_for_future[session_id].put(result)

def _put_exception(self, session_id, exception, **kwargs):
	result = dict(success=False, return_value=None, exception=exception, traceback=self._traceback(), cancelled=False)
	for key in result:
		if key in kwargs:
			result[key] = kwargs.pop(key)
	result["data"] = kwargs

	self._result_queues_for_future[session_id].put(result)

def _make_signal(self, session_id, **kwargs):
	cancel = False
	if "cancel" in kwargs:
		cancel = kwargs.pop("cancel")

	signal = \
	{
		"session_id": session_id,
		"type": "signal",
		"cancel": cancel,
		"data": kwargs
	}

	self._recved_signals[session_id].put(signal)
	
def _recv_signal(self, session_id):
	signal = self._recved_signals[session_id].get()

	if "debug" in signal["data"]:
		print("recved signal:", signal["data"]["debug"])

	return signal

def _recv_response(self, session_id):
	response = self._recved_responses[session_id].get()

	if "debug" in response["data"]:
		print("recved response:", response["data"]["debug"])

	return response

def _recving_loop(self):
	rear_binary = b''
	while True:
		while len(rear_binary) < 4:
			try:
				rear_binary += self._connection.recv(8192*1024)
			except:
				return
		var_length = int.from_bytes(rear_binary[:4], byteorder="little", signed=False)
		var_bytes = rear_binary[4:4+var_length]
		rear_binary = rear_binary[4+var_length:]
		if len(var_bytes) < var_length:
			delta_length = var_length - len(var_bytes)
			while len(rear_binary) < delta_length:
				try:
					rear_binary += self._connection.recv(8192*1024)
				except:
					self.close()
					return
			var_bytes += rear_binary[:delta_length]
			rear_binary = rear_binary[delta_length:]
		var = pickle.loads(var_bytes)
		if isinstance(var, dict) and ("type" in var) and ("session_id" in var):
			session_id = var["session_id"]
			if var["type"] == "request":
				if "debug" in var["data"]:
					print(var["data"]["debug"])

				if var["func"] in complex_request:
					self._complex_threads_pool.submit(eval("self._process_" + var["func"]), var)
				else:
					if var["func"] in ["_write_to_file", "_close_file"]:
						func = eval("self._process_" + var["func"])
						func(var)
					else:
						self._simple_threads_pool.submit(eval("self._process_" + var["func"]), var)
			elif var["type"] == "response":
				self._recved_responses[session_id].put(var)
			elif var["type"] == "signal":
				self._recved_signals[session_id].put(var)
		else:
			print("Recved not supported var:", var)


import socket
import threading
import os
import sys
from concurrent.futures import ThreadPoolExecutor

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _Node import Node
from _NodeInternalClasses import OrderedDict, CloseableQueue
from _utils import file_size

class Server:
	class Queues:
		def __init__(self):
			self._queue_dict = {}

		def __getitem__(self, name):
			if name not in self._queue_dict:
				self._queue_dict[name] = CloseableQueue()

			return self._queue_dict[name]

		def __delitem__(self, name):
			try:
				self._queue_dict[name].close()
			except:
				pass
			del self._queue_dict[name]

		def __len__(self):
			return len(self._queue_dict)

		def __contains__(self, name):
			return self._queue_dict.__contains__(name)

		def keys(self):
			return self._queue_dict.keys()

		def values(self):
			return self._queue_dict.values()

		def items(self):
			return self._queue_dict.items()

		def __iter__(self):
			return self._queue_dict.__iter__()

		def clear(self):
			for name in self._queue_dict:
				try:
					self._queue_dict[name].close()
				except:
					pass

			self._queue_dict.clear()

	def __init__(self, ip = None, port = None):
		self.connect_results = OrderedDict()
		self.disconnect_results = OrderedDict()
		self._callback_threads_pool = ThreadPoolExecutor(max_workers=5)

		self._onConnect = lambda client : None
		self._onDisconnect = lambda address : None

		self._connected_nodes = OrderedDict()
		self._global_vars = {}
		self._queue = CloseableQueue()
		self._queues = Server.Queues()

		self._continue_accept = False
		self._accept_thread = None
		
		self.start(ip, port)

	def set_connect_callback(self, func):
		self._onConnect = func

	def set_disconnect_callback(self, func):
		self._onDisconnect = func

	def start(self, ip=None, port=None):
		if self._continue_accept:
			return

		self._connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

		if ip == None:
			ip = [a for a in os.popen('route print').readlines() if ' 0.0.0.0 ' in a][0].split()[-2]
		if port == None:
			port = 0
		self._connection.bind((ip, port))
		self._connection.listen(5)
		
		self._continue_accept = True
		if self._accept_thread == None or not self._accept_thread.is_alive():
			self._accept_thread = threading.Thread(target=self._start, daemon=True)
			self._accept_thread.start()

	def close(self):
		self._continue_accept = False
		try:
			self._queues.clear()

			keys = self._connected_nodes.keys()
			for address in keys:
				try:
					self._connected_nodes[address].close()
				except:
					pass
		except:
			pass

		try:
			self._connection.close()
		except:
			pass
		
		self.connect_results = OrderedDict()
		self.disconnect_results = OrderedDict()
		self._callback_threads_pool.shutdown()
		self._callback_threads_pool = ThreadPoolExecutor(max_workers=5)

		self._connected_nodes = OrderedDict()
		self._global_vars = {}
		self._queues = Server.Queues()

		if self._accept_thread != None and self._accept_thread.is_alive():
			self._accept_thread.join()
		self._accept_thread = None
		self._connected_nodes = OrderedDict()

	@property	
	def address(self):
		if not self._continue_accept:
			raise ConnectionError("Connection is closed.")

		return self._connection.getsockname()

	@property
	def is_closed(self):
		return not self._continue_accept

	@property
	def clients(self):
		return self._connected_nodes

	@property
	def queues(self):
		return self._queues

	@property
	def globals(self):
		return self
	
	def __del__(self):
		try:
			self.close()
		except:
			pass

	def __getitem__(self, name):
		return self._global_vars[name]

	def __setitem__(self, name, value):
		self._global_vars[name] = value

	def __delitem__(self, name):
		del self._global_vars[name]

	def __iter__(self):
		return self._global_vars.__iter__()

	def __len__(self):
		return self._global_vars.__len__()

	def __contains__(self, name):
		return self._global_vars.__contains__(name)

	def keys(self):
		return self._global_vars.keys()

	def values(self):
		return self._global_vars.values()

	def items(self):
		return self._global_vars.items()

	def clear(self):
		self._global_vars.clear()

	def pop(self, name):
		return self._global_vars.pop(name)

	def get(self, timeout=None, block=True):
		return self._queue.get(timeout=timeout, block=block)

	def put(self, value, timeout=None, block=True):
		self._queue.put(value, timeout=timeout, block=block)

	def qsize(self):
		return self._queue.qsize()

	def put_file_to_all(self, src_filename, dest_filename = None):
		if not os.path.isfile(src_filename):
			raise FileNotFoundError("File " + src_filename + " is not exists.")

		if dest_filename == None:
			dest_filename = os.path.basename(src_filename)

		all_clients = self._connected_nodes.keys()

		block_size = 8192*1024
		sent_size = 0
		src_file_size = file_size(src_filename)
		with open(src_filename, "rb") as file:
			while sent_size < src_file_size:
				data = file.read(block_size)
				for address in all_clients:
					try:
						client = self._connected_nodes[address]
						client._write_to_file(data, dest_filename)
					except:
						pass
				sent_size += len(data)

		for address in all_clients:
			try:
				client = self._connected_nodes[address]
				client.close_file()
			except:
				pass

	def put_folder_to_all(self, src_foldername, dest_foldername = None):
		if not os.path.isdir(src_foldername):
			raise FileNotFoundError("Folder " + src_foldername + " is not exists.")

		if dest_foldername == None:
			dest_foldername = os.path.basename(src_foldername)

		i = len(src_foldername)-1
		while src_foldername[i] in ['/', '\\']:
			i -= 1

		all_clients = self._connected_nodes.keys()
		block_size = 8192*1024
		for root, dirs, files in os.walk(src_foldername):
			for name in files:
				src_filename = os.path.join(root, name)
				dest_filename = dest_foldername + "/" + src_filename[i+2:]

				sent_size = 0
				src_file_size = file_size(src_filename)
				with open(src_filename, "rb") as file:
					while sent_size < src_file_size:
						data = file.read(block_size)
						for address in all_clients:
							try:
								client = self._connected_nodes[address]
								client._write_to_file(data, dest_filename)
							except:
								pass
						sent_size += len(data)

		for address in all_clients:
			try:
				client = self._connected_nodes[address]
				client._close_file()
			except:
				pass

	def hold_on(self):
		threading.Event().wait()

	def _start(self):
		while self._continue_accept:
			try:
				client_socket, address = self._connection.accept()
				self._connected_nodes[address] = Node(client_socket, self)
				self.connect_results[address] = self._callback_threads_pool.submit(self._onConnect, self._connected_nodes[address])
			except BaseException as e:
				break


import threading
import time
import sys
import os
import ctypes
import collections

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _CloseableQueue import CloseableQueueFactory
from _utils import eprint

CloseableQueue = CloseableQueueFactory()

class OrderedDict(collections.OrderedDict):
	def __init__(self, **kwargs):
		collections.OrderedDict.__init__(self, **kwargs)

	def __setitem__(self, name, value):
		if isinstance(name, int):
			collections.OrderedDict.__setitem__(self, list(self.keys())[name], value)
		else:
			collections.OrderedDict.__setitem__(self, name, value)

	def __getitem__(self, name):
		if isinstance(name, int):
			return collections.OrderedDict.__getitem__(self, list(self.keys())[name])
		else:
			return collections.OrderedDict.__getitem__(self, name)

	def __delitem__(self, name):
		if isinstance(name, int):
			collections.OrderedDict.__delitem__(self, list(self.keys())[name])
		else:
			collections.OrderedDict.__delitem__(self, name)

class QueueDict(dict):
	def __init__(self):
		dict.__init__(self)

	def __getitem__(self, name):
		if not dict.__contains__(self, name):
			dict.__setitem__(self, name, CloseableQueue())
		return dict.__getitem__(self, name)

	def clear(self):
		for name in self:
			try:
				self[name].close()
			except:
				pass
		dict.clear(self)

class Pipe:
	def __init__(self, node, name):
		self._node = node
		self._name = name

	def send(self, value):
		self._node._locals_pipe_send(self._name, value)

	def recv(self):
		return self._node._locals_pipe_recv(self._name)

	def close(self):
		return self._node._locals_pipe_close(self._name)

class Pipes:
	def __init__(self, node):
		self._node = node

	def __getitem__(self, name):
		return Pipe(self._node, name)

	def __delitem__(self, name):
		self._node._locals_pipes_delitem(name)

	def __len__(self):
		return self._node._locals_pipes_len()

	def __contains__(self, name):
		return self._node._locals_pipes_contains(name)

	def __iter__(self):
		return self._node._locals_pipes_iter()

	def keys(self):
		return self._node._locals_pipes_keys()

	def values(self):
		values = []
		for key in self.keys():
			values.append(self[key])

		return values

	def items(self):
		items = []
		for key in self.keys():
			items.append((key, self[key]))

		return items

	def clear(self):
		self._node._locals_pipes_clear()
	
class Queue:
	def __init__(self, node, name, private=False):
		self._node = node
		self._name = name
		self._is_private = private

	def get(self, timeout = None, block = True):
		return self._node._locals_queue_get(self._name, timeout=timeout, block=block, is_private=self._is_private)

	def put(self, value, timeout = None, block = True):
		self._node._locals_queue_put(self._name, value, timeout=timeout, block=block, is_private=self._is_private)

	def close(self):
		self._node._locals_queue_close(self._name, is_private=self._is_private)

	def qsize(self):
		return self._node._locals_queue_len(self._name, is_private=self._is_private)

	def __len__(self):
		return self._node._locals_queue_len(self._name, is_private=self._is_private)

class Queues:
	def __init__(self, node, private=False):
		self._node = node
		self._is_private = private

	def __getitem__(self, name):
		return Queue(self._node, name, private=self._is_private)

	def __delitem__(self, name):
		self._node._locals_queues_delitem(name, is_private=self._is_private)

	def __len__(self):
		return self._node._locals_queues_len(is_private=self._is_private)

	def __contains__(self, name):
		return self._node._locals_queues_contains(name, is_private=self._is_private)

	def keys(self):
		return self._node._locals_queues_keys(is_private=self._is_private)

	def values(self):
		values = []
		for key in self.keys():
			values.append(self[key])

		return values

	def items(self):
		items = []
		for key in self.keys():
			items.append((key, self[key]))
		return items

	def __iter__(self):
		return self._node._locals_queues_iter(is_private=self._is_private)

	def clear(self):
		self._node._locals_queues_clear(is_private=self._is_private)

class Globals:
	class Queue:
		def __init__(self, node, name):
			self._node = node
			self._name = name

		def get(self, timeout = None, block = True):
			return self._node._globals_queue_get(self._name, timeout=timeout, block=block)

		def put(self, value, timeout = None, block = True):
			self._node._globals_queue_put(self._name, value, timeout=timeout, block=block)

		def close(self):
			self._node._globals_queue_close(self._name)

		def qsize(self):
			return self._node._globals_queue_len(self._name)

		def __len__(self):
			return self._node._globals_queue_len(self._name)

	class Queues:
		def __init__(self, node):
			self._node = node

		def __getitem__(self, name):
			return Globals.Queue(self._node, name)

		def __delitem__(self, name):
			self._node._globals_queues_delitem(name)

		def __len__(self):
			return self._node._globals_queues_len()

		def __contains__(self, name):
			return self._node._globals_queues_contains(name)

		def keys(self):
			return self._node._globals_queues_keys()

		def values(self):
			values = []
			for key in self.keys():
				values.append(self[key])
			return values

		def items(self):
			items = []
			for key in self.keys():
				items.append((key, self[key]))
			return items

		def __iter__(self):
			return self._node._globals_queues_iter()

		def clear(self):
			self._node._globals_queues_clear()

	def __init__(self, node):
		self._node = node
		self._queues = Globals.Queues(node)

	def __setitem__(self, name, value):
		self._node._globals_setitem(name, value)

	def __getitem__(self, name):
		return self._node._globals_getitem(name)

	def __delitem__(self, name):
		return self._node._globals_delitem(name)

	def __iter__(self):
		return self._node._globals_iter()

	def __contains__(self, name):
		return self._node._globals_contains(name)

	def __len__(self):
		return self._node._globals_len()

	def clear(self):
		self._node._globals_clear()

	def keys(self):
		return self._node._globals_keys()

	def values(self):
		return self._node._globals_values()

	def items(self):
		return self._node._globals_items()

	def pop(self, name):
		return self._node._globals_pop(name)

	def get(self, timeout=None, block=True):
		return self._node._globals_get(timeout=timeout, block=block)

	def put(self, value, timeout=None, block=True):
		self._node._globals_put(value, timeout=timeout, block=block)

	def qsize(self):
		return self._node._globals_qsize()

	@property
	def queues(self):
		return self._queues

class AutoSendBuffer:
	def __init__(self, node, session_id, timeout=0.1, quiet=False):
		self._node = node
		self._session_id = session_id
		self._buffer_lock = threading.Lock()
		self._buffer = b''
		self._quiet = quiet
		self._timeout = timeout
		self._last_send_time = time.time()
		self._continue_send = True
		self._sending_thread = threading.Thread(target=self._timeout_send)
		self._sending_thread.start()

	def __del__(self):
		try:
			self.stop()
		except:
			pass

	def append(self, byte):
		with self._buffer_lock:
			self._buffer += byte

		if byte == b"\n" or len(self._buffer) >= 512:
			self.send()

	def send(self):
		with self._buffer_lock:
			if len(self._buffer) != 0:
				content = self._buffer.decode("utf-8")
				if not self._quiet:
					print(content, end="", flush=True)
				self._node._respond_ok(self._session_id, end=False, stdout=content)
				self._buffer = b''
				self._last_send_time = time.time()

	def stop(self):
		self._continue_send = False
		self._sending_thread.join()
		self.send()

	def _timeout_send(self):
		while self._continue_send:
			if time.time()-self._last_send_time >= self._timeout:
				self.send()
			time.sleep(1E-3)

class Thread(threading.Thread): 
	def __init__(self, target, args=(), kwargs={}): 
		threading.Thread.__init__(self, target=target, args=args, kwargs=kwargs) 
		
	def __get_id(self): 
		if hasattr(self, '_thread_id'): 
			return self._thread_id 
		for id, thread in threading._active.items(): 
			if thread is self: 
				return id

	def kill(self):
		thread_id = self.__get_id()
		res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, ctypes.py_object(SystemExit))
		if res > 1:
			ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
			print('Exception raise failure')

class Future:
	def __init__(self, node, session_id):
		self._node = node
		self._session_id = session_id
		self._cancelled = False
		self._done = None
		self._result = None
		self._monitor_done_thread = None
		self._done_callbacks = []

	def cancel(self):
		if self.done():
			return False

		self._node._stop_send = True
		self._node._send_signal(self._session_id, cancel=True)
		self._result = self._node._result_queues_for_future[self._session_id].get()
		self._cancelled = True
		self._done = True
		return True

	def cancelled(self):
		return self._cancelled

	def done(self):
		if self._done != None:
			return self._done
		else:
			return (len(self._node._result_queues_for_future[self._session_id]) != 0)

	def result(self):
		if self._result == None:
			self._result = self._node._result_queues_for_future[self._session_id].get()

		self._done = True
		if self._result["success"]:
			return self._result["return_value"]
		else:
			eprint(self._result["traceback"])
			raise self._result["exception"]

	def exception(self):
		if self._result == None:
			self._result = self._node._result_queues_for_future[self._session_id].get()

		self._done = True
		return self._result["exception"]

	def _get_result(self):
		if self._result == None:
			self._result = self._node._result_queues_for_future[self._session_id].get()

		for func in self._done_callbacks:
			func(self)

	def add_done_callback(self, func):
		self._done_callbacks.append(func)
		
		if self._monitor_done_thread == None:
			self._monitor_done_thread = threading.Thread(target=self._get_result)
			self._monitor_done_thread.start()

	def remove_done_callback(self, func):
		try:
			self._done_callbacks.remove(func)
		except:
			pass


import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _utils import eprint

def __getitem__(self, name):
	if self._server == None:
		return self._local_vars[name]
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def __setitem__(self, name, value):
	if self._server == None:
		self._local_vars[name] = value
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, value=value)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def __delitem__(self, name):
	if self._server == None:
		del self._local_vars[name]
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def __iter__(self):
	if self._server == None:
		return self._local_vars.__iter__()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["iter"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def __contains__(self, name):
	if self._server == None:
		return self._local_vars.__contains__(name)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["contains"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def __len__(self):
	if self._server == None:
		return self._local_vars.__len__()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		if response["success"]:
			return response["data"]["len"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def clear(self):
	if self._server == None:
		self._local_vars.clear()
	else:
		session_id = self._get_session_id()

		self._request(session_id)
		response = self._recv_response(session_id)
		
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def keys(self):
	if self._server == None:
		return self._local_vars.keys()
	else:
		session_id = self._get_session_id()

		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["keys"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def values(self):
	if self._server == None:
		return self._local_vars.values()
	else:
		session_id = self._get_session_id()

		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["values"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def items(self):
	if self._server == None:
		return self._local_vars.items()
	else:
		session_id = self._get_session_id()

		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["items"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def pop(self, name):
	if self._server == None:
		return self._local_vars.pop(name)
	else:
		session_id = self._get_session_id()

		self._request(session_id, name=name)
		response = self._recv_response(session_id)

		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process___getitem__(self, request):
	try:
		self._respond_ok(request["session_id"], value=self._local_vars[request["data"]["name"]])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process___setitem__(self, request):
	try:
		self._local_vars[request["data"]["name"]] = request["data"]["value"]
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process___delitem__(self, request):
	try:
		del self._local_vars[request["data"]["name"]]
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process___iter__(self, request):
	try:
		self._respond_ok(request["session_id"], iter=self._local_vars.__iter__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process___contains__(self, request):
	try:
		self._respond_ok(request["session_id"], contains=self._local_vars.__contains__(request["data"]["name"]))
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process___len__(self, request):
	try:
		self._respond_ok(request["session_id"], len=self._local_vars.__len__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process_clear(self, request):
	try:
		self._local_vars.clear()
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process_keys(self, request):
	try:
		self._respond_ok(request["session_id"], keys=self._local_vars.keys())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process_values(self, request):
	try:
		self._respond_ok(request["session_id"], values=self._local_vars.values())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process_items(self, request):
	try:
		self._respond_ok(request["session_id"], items=self._local_vars.items())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _process_pop(self, request):
	try:
		self._respond_ok(request["session_id"], value=self._local_vars.pop(request["data"]["name"]))
	except BaseException as e:
		self._respond_exception(request["session_id"], e)


import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _utils import eprint

def _globals_get(self, timeout=None, block=True):
	if self._server != None:
		return self._server.get(timeout=timeout, block=block)
	else:
		session_id = self._get_session_id()
		self._request(session_id, timeout=timeout, block=block)
		response = self._recv_response(session_id)

		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_get(self, request):
	try:
		value = self._server.get(timeout=request["data"]["timeout"], block=request["block"])
		self._respond_ok(request["session_id"], value=value)
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_put(self, value, timeout=None, block=True):
	if self._server != None:
		return self._server.put(value, timeout=timeout, block=block)
	else:
		session_id = self._get_session_id()
		self._request(session_id, value=value, timeout=timeout, block=block)
		response = self._recv_response(session_id)

		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_put(self, request):
	try:
		self._server.put(request["data"]["value"], timeout=request["data"]["timeout"], block=request["block"])
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_qsize(self, timeout=None, block=True):
	if self._server != None:
		return self._server.qsize()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)

		if response["success"]:
			return response["data"]["qsize"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_qsize(self, request):
	try:
		qsize = self._server.qsize()
		self._respond_ok(request["session_id"], qsize=qsize)
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queue_get(self, name, timeout=None, block=True):
	if self._server != None:
		return self._server.queues[name].get(timeout=timeout, block=block)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, timeout=timeout, block=block)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["value"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queue_get(self, request):
	try:
		value = self._server.queues[request["data"]["name"]].get(timeout=request["data"]["timeout"], block=request["block"])
		self._respond_ok(request["session_id"], value=value)
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queue_put(self, name, value, timeout = None, block = True):
	if self._server != None:
		self._server.queues[name].put(value, timeout = timeout, block=block)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name, value=value, timeout = timeout)
		response = self._recv_response(session_id)
		
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queue_put(self, request):
	try:
		self._server.queues[request["data"]["name"]].put(request["data"]["value"], timeout=request["data"]["timeout"], block=request["block"])
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queue_len(self, name):
	if self._server != None:
		if name not in self._server.queues:
			return 0
		else:
			return self._server.queues[name].qsize()
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["len"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queue_len(self, request):
	try:
		if request["data"]["name"] not in self._server.queues:
			self._respond_ok(request["session_id"], len=0)
		else:
			self._respond_ok(request["session_id"], len=self._server.queues[request["data"]["name"]].qsize())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queue_close(self, name):
	if self._server != None:
		if name in self._server.queues:
			try:
				return self._server.queues[name].close()
			except:
				pass
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["len"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queue_close(self, request):
	try:
		name = request["data"]["name"]
		if name in self._server.queues:
			try:
				self._server.queues[name].close()
			except:
				pass
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queues_delitem(self, name):
	if self._server != None:
		del self._server.queues[name]
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queues_delitem(self, request):
	try:
		del self._server.queues[request["data"]["name"]]
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queues_len(self):
	if self._server != None:
		return len(self._server.queues)
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["len"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queues_len(self, request):
	try:
		self._respond_ok(request["session_id"], len=self._server.queues.__len__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queues_contains(self, name):
	if self._server != None:
		return self._server.queues.__contains__(name)
	else:
		session_id = self._get_session_id()
		self._request(session_id, name=name)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["contains"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queues_contains(self, request):
	try:
		self._respond_ok(request["session_id"], contains=self._server.queues.__contains__(request["data"]["name"]))
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queues_keys(self):
	if self._server != None:
		return self._server.queues.keys()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["keys"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queues_keys(self, request):
	try:
		self._respond_ok(request["session_id"], keys=self._server.queues.keys())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queues_iter(self):
	if self._server != None:
		return self._server.queues.__iter__()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if response["success"]:
			return response["data"]["iter"]
		else:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queues_iter(self, request):
	try:
		self._respond_ok(request["session_id"], iter=self._server.queues.__iter__())
	except BaseException as e:
		self._respond_exception(request["session_id"], e)

def _globals_queues_clear(self):
	if self._server != None:
		self._server.queues.clear()
	else:
		session_id = self._get_session_id()
		self._request(session_id)
		response = self._recv_response(session_id)
		
		if not response["success"]:
			eprint(response["traceback"])
			raise response["exception"]

def _process__globals_queues_clear(self, request):
	try:
		self._server.queues.clear()
		self._respond_ok(request["session_id"])
	except BaseException as e:
		self._respond_exception(request["session_id"], e)



import socket
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from _Node import Node

class Client(Node):
	def __init__(self, ip = None, port = None):
		self._is_Node_init = False
		if ip == None:
			ip = [a for a in os.popen('route print').readlines() if ' 0.0.0.0 ' in a][0].split()[-2]
		if port == None:
			port = 0

		self._ip = ip
		self._port = port

		self._connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self._connection.bind((ip, port))
		self._address = self._connection.getsockname()

	def connect(self, ip, port = None):
		if self._is_Node_init and not self.is_closed:
			self.close()

		server_address = ip
		if port != None:
			server_address = (ip, port)

		if self._connection == None:
			self._connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self._connection.bind((self._ip, self._port))
			self._address = self._connection.getsockname()

		self._connection.connect(server_address)
		Node.__init__(self, self._connection)